import torch
import torch.nn as nn
import torch.nn.functional as F

class EncBlock(torch.nn.Module):
    def __init__(self, in_dim, out_dim, kernel_size, stride, padding, pool=False, bn=False, dropout=0.0, actv_layer=torch.nn.SiLU(), skip_connection=False):
        super().__init__()
        self.pool = pool
        self.bn = bn
        self.skip_connection = skip_connection

        modules = [torch.nn.Conv2d(in_dim, out_dim, kernel_size, stride, padding)]
        if pool:
            modules.append(torch.nn.MaxPool2d(kernel_size=2, stride=2))
        if bn:
            modules.append(torch.nn.BatchNorm2d(out_dim))
        if dropout > 0.0:
            modules.append(torch.nn.Dropout2d(dropout))
        if actv_layer is not None:
            modules.append(actv_layer)
        self.net = torch.nn.Sequential(*modules)
    
    def forward(self, x):
        y = self.net(x)
        if self.skip_connection:
            if self.pool:
                skip = torch.nn.functional.avg_pool2d(x, kernel_size=2, stride=2)
            else:
                skip = x
            y += skip
        return y

class DecBlock(torch.nn.Module):
    def __init__(self, in_dim, out_dim, kernel_size, stride=1, padding=0, upsample=False, bn=False, dropout=0.0, actv_layer=torch.nn.SiLU(), skip_connection=False):
        super().__init__()
        self.upsample = upsample
        self.bn = bn
        self.skip_connection = skip_connection

        modules = [torch.nn.ConvTranspose2d(in_dim, out_dim, kernel_size, stride, padding)]
        if upsample:
            modules.append(torch.nn.Upsample(scale_factor=2, mode='bilinear'))
        if bn:
            modules.append(torch.nn.BatchNorm2d(out_dim))
        if dropout > 0.0:
            modules.append(torch.nn.Dropout2d(dropout))
        if actv_layer is not None:
            modules.append(actv_layer)
        self.net = torch.nn.Sequential(*modules)
    
    def forward(self, x):
        y = self.net(x)
        if self.skip_connection:
            if self.upsample:
                skip = torch.nn.functional.interpolate(x, scale_factor=2, mode='bilinear')
            else:
                skip = x
            y += skip
        return y

        
class ConvResBlock(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, gn_groups: int, actv_layer: nn.Module = nn.SiLU(), dropout: float = 0.0, residual: bool = True, scale: float = 1.0, num_layers: int = 2):
        super().__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.do_gn = gn_groups > 0
        self.do_actv = actv_layer is not None
        self.do_dropout = dropout > 0.0
        self.do_residual = residual
        self.scale = scale
        self.num_layers = num_layers

        assert num_layers > 0, "num_layers must be greater than 0"
        self.convs = nn.ModuleList([nn.Conv2d(in_channels, out_channels, kernel_size=3, padding=1)])
        for _ in range(num_layers-1):
            self.convs.append(nn.Conv2d(out_channels, out_channels, kernel_size=3, padding=1))

        if self.do_gn:
            self.groupnorms = nn.ModuleList([nn.GroupNorm(gn_groups, in_channels)])
            for _ in range(num_layers-1):
                self.groupnorms.append(nn.GroupNorm(gn_groups, out_channels))
        
        if self.do_dropout:
            self.dropout = nn.Dropout2d(dropout)

        if self.do_actv:
            self.actv_fn = actv_layer

        if self.do_residual:
            if in_channels == out_channels:
                self.residual_layer = nn.Identity()
            else:
                self.residual_layer = nn.Conv2d(in_channels, out_channels, kernel_size=1, padding=0)

        if scale < 1.0:
            stride = int(1 / scale)
            self.down = nn.Conv2d(in_channels, in_channels, kernel_size=3, stride=stride, padding=0)
        elif scale > 1.0:
            self.up = nn.Upsample(scale_factor=scale)


    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # x: (Batch_Size, Channel, Height, Width)

        if self.scale < 1.0:
            # Downsample input
            x = self.down(x)

        # Save input for residual connection
        if self.do_residual:
            residual = x
        
        # Apply layers
        for i in range(self.num_layers):
            # Groupnorm
            if self.do_gn:
                x = self.groupnorms[i](x)
            # Activation
            if self.do_actv:
                x = self.actv_fn(x)
            # Dropout
            if self.do_dropout:
                x = self.dropout(x)
            # Convolution
            x = self.convs[i](x)

        if self.do_residual:
            # Add residual connection
            x = x + self.residual_layer(residual)

        if self.scale > 1.0:
            # Upsample
            x = self.up(x)

        return x


import torch
import torch.nn as nn
import torchvision.transforms as transforms
import torchvision.datasets as datasets

from joetorch.datasets.dataset import PreloadedDataset

def FashionMNIST(root, split, val_ratio=0.1, normalize=False, transform=None, dtype='float32', device='cpu', download=True):
    assert normalize is False, "Normalization is not supported for FashionMNIST"
    return MNIST(root, split, val_ratio, normalize, transform, dtype, device, download, fashion=True)

def MNIST(
        root, 
        split, 
        val_ratio=0.0,
        normalize=True,
        augment=False,
        transform=None, 
        dtype=torch.float32,
        device=torch.device('cpu'),
        download=True,
        fashion=False,
    ):
    # Load data
    assert split in ['train', 'val', 'test']

    train = split in ['train', 'val'] # True for train and val, False for test
    loading_transform = transforms.Compose([transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))]) if normalize else transforms.ToTensor()
    if fashion:
        dataset = datasets.FashionMNIST(root=root, train=train, transform=loading_transform, download=download)
    else:
        dataset = datasets.MNIST(root=root, train=train, transform=loading_transform, download=download)

    fill_value = dataset[0][0].min().item()

    if transform is None:
        transform = []
    if type(transform) is not list:
        transform = [transform]
    if augment:
        transform = [transforms.RandomAffine(degrees=15, translate=(0.1, 0.1), scale=(0.9, 1.1), shear=10, fill=fill_value)] + transform
    if len(transform) > 0:
        transform = transforms.Compose(transform)
    else:
        transform = None

    if split == 'train':
        # Build train dataset
        n_train = int(len(dataset) * (1 - val_ratio))
        dataset = torch.utils.data.Subset(dataset, range(0, n_train))
        dataset = PreloadedDataset.from_dataset(dataset, transform, device, use_tqdm=True)
    
    elif split == 'val':
        # Build val dataset
        n_val = int(len(dataset) * val_ratio)
        dataset = torch.utils.data.Subset(dataset, range(len(dataset) - n_val, len(dataset)))
        dataset = PreloadedDataset.from_dataset(dataset, transform, device, use_tqdm=True)
    
    elif split == 'test':
        dataset = PreloadedDataset.from_dataset(dataset, transform, device, use_tqdm=True)

    dataset = dataset.to(dtype).to(device)

    return dataset


import os
import torch
from joetorch.datasets.dataset import PreloadedDataset
from torch.utils.tensorboard import SummaryWriter
from typing import Optional, Union, List, Dict, Tuple
from torch.utils.data import DataLoader
import numpy as np
from tqdm import tqdm


def train(
        model: torch.nn.Module,
        train_dataset: PreloadedDataset,
        optimiser: torch.optim.Optimizer,
        num_epochs: int,
        batch_size: int,
        val_dataset: Optional[PreloadedDataset] = None,
        writer: Optional[SummaryWriter] = None,
        compute_dtype: str = 'float32',
        epoch_hyperparams: Dict[str, torch.Tensor] = {},
        save_dir: Optional[str] = None,
        save_metric: Tuple[str, str, str] = ('minimise', 'val', 'loss'),
        save_every: Optional[int] = None,
        save_last: bool = False,
        **kwargs,
):
    """
    Train a model for a given number of epochs.

    Args:
        model: The model to train.
        train_dataset: The training dataset.
        val_dataset: The validation dataset.
        optimiser: The optimiser to use.
        num_epochs: The number of epochs to train for.
        batch_size: The batch size.
        writer: The writer to use for logging.
        mixed_precision: Whether to use mixed precision.
        epoch_hyperparams: The hyperparameters to use for each epoch.
        save_dir: The directory to save the model.
        save_metric: Tuple of ('minimise'/'maximise', 'train'/'val', {metric}).
    """
    assert save_metric[0] in ['minimise', 'maximise']
    assert save_metric[1] in ['train', 'val']

    compute_dtype = getattr(torch, str(compute_dtype).split('.')[-1])
    compute_device = next(model.parameters()).device
    train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True)
    val_loader = None if val_dataset is None else DataLoader(val_dataset, batch_size=batch_size, shuffle=False)

    loop = tqdm(range(num_epochs), desc='Epochs', leave=False)
    best_metric = torch.inf if save_metric[0] == 'minimise' else -torch.inf
    for epoch in loop:

        # ========================== HYPERPARAMETER UPDATE ==========================

        epoch_loss_args = {}
        for key, value in epoch_hyperparams.items():
            if key == 'lr':
                for param_group in optimiser.param_groups:
                    param_group['lr'] = value[epoch]
            elif key == 'wd':
                for param_group in optimiser.param_groups:
                    if param_group['weight_decay'] != 0.0:
                        param_group['weight_decay'] = value[epoch]
            else:
                if isinstance(value, torch.Tensor) or isinstance(value, np.ndarray) or isinstance(value, list):
                    epoch_loss_args[key] = value[epoch]
                else:
                    epoch_loss_args[key] = value
                    
        # ============================ TRAINING ============================
        if train_dataset.transform is not None:
            train_dataset.update_transformed_images()

        model.train()
        epoch_train_metrics = {}
        for batch in train_loader:
            optimiser.zero_grad()
            with torch.autocast(device_type=compute_device.type, dtype=compute_dtype, enabled=compute_dtype is not None):
                train_metrics = model.loss(batch, **epoch_loss_args)
                train_metrics['loss'].backward()
                optimiser.step()
            
            for key, value in train_metrics.items():
                if key not in epoch_train_metrics:
                    epoch_train_metrics[key] = []
                epoch_train_metrics[key].append(value.detach().cpu().item())

        for key, value in epoch_train_metrics.items():
            epoch_train_metrics[key] = np.mean(value)

        if writer is not None:
            for key, value in epoch_train_metrics.items():
                writer.add_scalar(f'train/{key}', value, epoch)
            
        postfix = {key: round(value.item(), 3) for key, value in epoch_train_metrics.items()}
        # ============================ VALIDATION ============================

        if val_loader is not None:
            if val_dataset.transform is not None:
                val_dataset.update_transformed_images()

            model.eval()
            epoch_val_metrics = {}
            for batch in val_loader:
                with torch.no_grad():
                    with torch.autocast(device_type=compute_device.type, dtype=compute_dtype, enabled=compute_dtype is not None):
                        val_metrics = model.loss(batch, **epoch_loss_args)

                for key, value in val_metrics.items():
                    if key not in epoch_val_metrics:
                        epoch_val_metrics[key] = []
                    epoch_val_metrics[key].append(value.detach().cpu().item())
            
            # ============================ LOGGING ============================

            # average metrics over the epoch
            for key, value in epoch_val_metrics.items():
                epoch_val_metrics[key] = np.mean(value)

            if writer is not None:
                for key, value in epoch_val_metrics.items():
                    writer.add_scalar(f'val/{key}', value, epoch)
            
            for val_key, val_value in epoch_val_metrics.items():
                val_value = round(val_value.item(), 3)
                if val_key in postfix:
                    train_value = postfix[val_key]
                    postfix[val_key] = (train_value, val_value)
                else:
                    postfix[val_key] = val_value

        loop.set_postfix(postfix)

        # ============================ SAVING ============================

        if save_dir is not None:
            if not os.path.exists(save_dir):
                os.makedirs(save_dir)
            metrics = epoch_val_metrics if save_metric[1] == 'val' else epoch_train_metrics
            if (save_metric[0] == 'minimise' and metrics[save_metric[2]] < best_metric) or (save_metric[0] == 'maximise' and metrics[save_metric[2]] > best_metric):
                best_metric = metrics[save_metric[2]]
                for f in os.listdir(save_dir):
                    if 'best' in f:
                        os.remove(os.path.join(save_dir, f))
                torch.save(model.state_dict(), os.path.join(save_dir, f'best_{save_metric[1]}_{save_metric[2]}_{epoch}.pth'))
            if save_every is not None and epoch % save_every == 0:
                torch.save(model.state_dict(), os.path.join(save_dir, f'checkpoint_{epoch}.pth'))
            elif save_last:
                for f in os.listdir(save_dir):
                    if 'recent' in f:
                        os.remove(os.path.join(save_dir, f))
                torch.save(model.state_dict(), os.path.join(save_dir, f'recent_{epoch}.pth'))


import torch

class MLP(torch.nn.Module):
    def __init__(self, in_dim: int, out_dim: int, hidden_dim: int=1024, num_hidden: int=2, hidden_actv_module: torch.nn.Module=torch.nn.SiLU(), out_actv_module: torch.nn.Module=None):
        super().__init__()

        layers = []
        in_features = in_dim
        
        # Add hidden layers
        for _ in range(num_hidden):
            layers.append(torch.nn.Linear(in_features, hidden_dim))
            if hidden_actv_module is not None:
                layers.append(hidden_actv_module)
            in_features = hidden_dim

        # Add output layer
        layers.append(torch.nn.Linear(in_features, out_dim))
        if out_actv_module is not None:
            layers.append(out_actv_module)

        self.in_dim = in_dim
        self.out_dim = out_dim
        self.net = torch.nn.Sequential(*layers)

    def forward(self, x: torch.Tensor):
        return self.net(x)


import torch

def get_optimiser(model, optim='AdamW', betas=(0.9, 0.999), momentum=0.9, lr=99999.9, wd=99999.9):

    non_decay_parameters = []
    decay_parameters = []   
    for n, p in model.named_parameters():
        if 'bias' in n or 'bn' in n:
            non_decay_parameters.append(p)
        else:
            decay_parameters.append(p)
    non_decay_parameters = [{'params': non_decay_parameters, 'weight_decay': 0.0}]
    decay_parameters = [{'params': decay_parameters}]

    if optim == 'AdamW':
        optimiser = torch.optim.AdamW(decay_parameters + non_decay_parameters, lr=lr, weight_decay=wd, betas=betas)
    elif optim == 'SGD':
        optimiser = torch.optim.SGD(decay_parameters + non_decay_parameters, lr=lr, weight_decay=wd, momentum=momentum)
    
    return optimiser

def apply_weight_decay_with_prior(model, prior, weight_decay, lr, fisher=None):
    """
    Apply weight decay to the model parameters towards the given state dict.

    Args:
        model: The model whose parameters will be decayed.
        prior: A state dict of model parameters to which model parameters will be decayed.
        weight_decay: The weight decay value to be applied.
        lr: The learning rate value to be applied.
        fisher: An optional state dict of model parameters which scales weight decay updates by parameter-wise importance.
    """
    
    with torch.no_grad():
        for name, param in model.named_parameters():
            if name in prior:
                if fisher is None:
                    param.data += (prior[name].data - param.data) * weight_decay * lr
                else:
                    param.data += (prior[name].data - param.data) * weight_decay * lr * fisher[name]